// BackPython.cpp : implementation file
//

#include "stdafx.h"
#include "UruPython3.h"
#include "BackPython.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


CBackPython *g_backPython = NULL;


const int WM_PYTHON_START			= WM_PYTHON_QUIT + 1;
const int WM_PYTHON_STOP			= WM_PYTHON_QUIT + 2;
const int WM_PYTHON_BATCH			= WM_PYTHON_QUIT + 3;


/////////////////////////////////////////////////////////////////////////////
// CBackPython

IMPLEMENT_DYNCREATE(CBackPython, CWinThread)

CBackPython::CBackPython()
{
	m_pythonObj = NULL;

	m_dataBatch.done = TRUE;
	m_dataBatch.curFileNbr = 0;
	m_dataBatch.totalFile = 0;
	m_dataBatch.curFilename.Empty();
	m_dataBatch.filenameToDo.Empty();
	m_dataBatch.sourceDir.Empty();
	m_dataBatch.isErr = FALSE;
	m_dataBatch.errMsg.Empty();
	m_dataBatch.decompyle = TRUE;

	m_stopBatch = FALSE;
}

CBackPython::~CBackPython()
{
}

BOOL CBackPython::InitInstance()
{
	// TODO:  perform and per-thread initialization here
	return TRUE;
}

int CBackPython::ExitInstance()
{
	// Terminate Python object before leaving
	StopPythonMsg(0, 0);

	return CWinThread::ExitInstance();
}

BEGIN_MESSAGE_MAP(CBackPython, CWinThread)
	//{{AFX_MSG_MAP(CBackPython)
	//}}AFX_MSG_MAP
	ON_THREAD_MESSAGE(WM_PYTHON_QUIT, QuitThread)
	ON_THREAD_MESSAGE(WM_PYTHON_START, StartPythonMsg)
	ON_THREAD_MESSAGE(WM_PYTHON_STOP, StopPythonMsg)
	ON_THREAD_MESSAGE(WM_PYTHON_BATCH, BatchMsg)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CBackPython message handlers

void CBackPython::QuitThread( UINT wParam, LONG lParam) {
	::PostQuitMessage(0);
}


BOOL CBackPython::IsPythonStarted()
{
	return(m_pythonObj != NULL);
}

// Start Python object in second thread an return FALSE in case of error.  The object is started in
// the second thread.
BOOL CBackPython::StartPython(CString &errMsg)
{
	m_EventCmdDone.ResetEvent();
	PostThreadMessage(WM_PYTHON_START, 0, 0);

	// Wait for end of command.
	CSingleLock myLock(&m_EventCmdDone);
	myLock.Lock();

	errMsg = m_errStart;
	return m_resultStart;
}

void CBackPython::StartPythonMsg( UINT wParam, LONG lParam) {
	m_resultStart = TRUE;
	m_errStart.Empty();

	if (!m_pythonObj) {
		m_pythonObj = new CPython;
		if (!m_pythonObj) {
			m_resultStart = FALSE;
			m_errStart = _T("Error with 'new CPython' (memory error?)");;
		} else {
			m_resultStart = m_pythonObj->OpenLink(m_errStart);
			if (m_resultStart) {
				// Empty stderr buffer (sometime filled with an error message if python.pak file not present)
				CString msg = m_pythonObj->GetOutput();
			}
		}
	}

	// Set the event to signal the end of the operation
	m_EventCmdDone.SetEvent();
}

// Stop python interpreter (usefull to update python DLL)
void CBackPython::StopPython()
{
	StopBatch();

	m_EventCmdDone.ResetEvent();
	PostThreadMessage(WM_PYTHON_STOP, 0, 0);

	// Wait for end of command.
	CSingleLock myLock(&m_EventCmdDone);
	myLock.Lock();
}


void CBackPython::StopPythonMsg( UINT wParam, LONG lParam)
{
	if (m_pythonObj) {
		m_pythonObj->CloseLink();
		delete m_pythonObj;
		m_pythonObj = NULL;
	}

	// Set the event to signal the end of the operation
	m_EventCmdDone.SetEvent();
}


// Stop the decompylation/compilation of PYC/PY file.  Waits until the current file is completed (may takes several
// minutes).
void CBackPython::StopBatch()
{
	m_lockDataBatch.Lock();
	BOOL isDone = m_dataBatch.done;
	if (!isDone)
		m_EventDecompyle.ResetEvent();		// Make sure other thread cannot set 'isDone' before I reset the event
	m_lockDataBatch.Unlock();

	if (!isDone) {
		m_stopBatch = TRUE;

		// Wait for end of command.
		CSingleLock myLock(&m_EventDecompyle);
		myLock.Lock();
	}
}

// Get status on the current decompylation/compilation session
void CBackPython::GetBatchStatus(CStatBatch &data)
{
	m_lockDataBatch.Lock();
	data = m_dataBatch;
	m_lockDataBatch.Unlock();
}

// Start decompylation/compilation process.  This function returns immediately.  Caller must use GetBatchStatus
// to monitor progress done by second thread.  Return FALSE in case of error (second thread not activated in
// this case).  'fileList' is a comma separeted list of file without extension.
BOOL CBackPython::StartBatch(CString sourceDir, CString fileList, BOOL checkDate, BOOL isDecompyle)
{
	if (!IsPythonStarted())
		return FALSE;

	StopBatch();

	// Check parameter
	sourceDir.TrimLeft();
	sourceDir.TrimRight();

	fileList.TrimLeft();
	fileList.TrimRight();

	if (fileList.IsEmpty())
		return FALSE;

	// Build the structure required by BatchMsg function
	CStatBatch data;
	data.done = FALSE;
	data.curFileNbr = 0;
	data.totalFile = CountComma(fileList) + 1;
	data.curFilename.Empty();
	data.filenameToDo = fileList;
	data.sourceDir = sourceDir;
	data.isErr = FALSE;
	data.errMsg.Empty();
	data.decompyle = isDecompyle;

	m_lockDataBatch.Lock();
	m_dataBatch = data;
	m_lockDataBatch.Unlock();

	// Start processing in other thread
	m_stopBatch = FALSE;
	PostThreadMessage(WM_PYTHON_BATCH, checkDate, 0);

	return TRUE;
}




// Do the decompylation of file based on the content of the m_dataBatch structure.  Stop when all
// file process or error on one file or m_stopBatch is TRUE.
void CBackPython::BatchMsg( UINT checkDate, LONG lParam) {
	// Works with a local copy of the structure
	m_lockDataBatch.Lock();
	CStatBatch data = m_dataBatch;
	m_lockDataBatch.Unlock();

	// Make sure sourceDir has a backslash
	if (!data.sourceDir.IsEmpty() && (data.sourceDir[data.sourceDir.GetLength() - 1] != _T('\\')))
		data.sourceDir += _T('\\');

	CString readFilename, writeFilename;

	while (1) {
		if (m_stopBatch) {
			data.isErr = TRUE;
			data.errMsg = _T("Stopped by operator");
			break;
		}

		if (data.filenameToDo.IsEmpty())
			break;

		// Extract next filename
		int posComma = data.filenameToDo.Find(_T(','));
		if (posComma >= 0)
			data.curFilename = data.filenameToDo.Left(posComma);
		else
			data.curFilename = data.filenameToDo;

		data.curFilename.TrimLeft();
		data.curFilename.TrimRight();

		// Update status for other thread
		data.curFileNbr++;
		m_lockDataBatch.Lock();
		m_dataBatch = data;
		m_lockDataBatch.Unlock();

		if (!data.curFilename.IsEmpty()) {
			readFilename = data.sourceDir + data.curFilename + _T(".pyc");
			writeFilename = data.sourceDir + data.curFilename + _T(".py");
			if (!data.decompyle) {
				CString tempStr = readFilename;
				readFilename = writeFilename;
				writeFilename = tempStr;
			}

			// Make sure source file exist
			BOOL processIt = TRUE;
			CTime dateSource;
			CFileFind finder;
			BOOL fileExist = finder.FindFile(readFilename);
			if (fileExist) {
				finder.FindNextFile();
				finder.GetLastWriteTime(dateSource);
			}
			finder.Close();

			if (!fileExist) {
				data.isErr = TRUE;
				data.errMsg = _T("File not found");
				break;
			}

			if (checkDate) {
				// Check if writeFilename is newer than readFilename
				CTime dateDest;
				fileExist = finder.FindFile(writeFilename);
				if (fileExist) {
					finder.FindNextFile();
					finder.GetLastWriteTime(dateDest);
				}
				finder.Close();

				if (fileExist && (dateDest > dateSource))
					processIt = FALSE;
			}

			if (processIt) {
				// Empty python stderr before trying to do the decompylation
				CString msgPython = m_pythonObj->GetOutput();
				if (data.decompyle) {
					// Decompile
					if (!m_pythonObj->Decompyle(readFilename, writeFilename, msgPython)) {
						data.isErr = TRUE;
						data.errMsg = msgPython;
						break;
					}
				} else {
					// Compile
					if (!m_pythonObj->Compile(readFilename, msgPython)) {
						data.isErr = TRUE;
						data.errMsg = msgPython;
						break;
					}
				}
			}
		}

		// Extract from the list of file the last filename processed 
		if (posComma >= 0)
			data.filenameToDo = data.filenameToDo.Mid(posComma + 1);
		else
			data.filenameToDo.Empty();

		data.filenameToDo.TrimLeft();
		data.filenameToDo.TrimRight();
	}

	// Set last bit of information for the other task (mainly, it's done)
	data.done = TRUE;
	m_lockDataBatch.Lock();
	m_dataBatch = data;
	m_EventDecompyle.SetEvent();
	m_lockDataBatch.Unlock();

}
