// DecompyleDiag.cpp : implementation file
//

#include "stdafx.h"
#include "UruPython3.h"
#include "DecompyleDiag.h"
#include "BackPython.h"
#include "FolderSelect.h"
#include "Progress.h"
#include "ErrDecomp.h"
#include "SelectFile.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CDecompyleDiag dialog


CDecompyleDiag::CDecompyleDiag(CWnd* pParent /*=NULL*/)
	: CDialog(CDecompyleDiag::IDD, pParent)
{
	//{{AFX_DATA_INIT(CDecompyleDiag)
	//}}AFX_DATA_INIT
}


void CDecompyleDiag::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CDecompyleDiag)
	//}}AFX_DATA_MAP
	DDX_Check(pDX, IDC_CHECKDATE, g_config.compareDate);
}


BEGIN_MESSAGE_MAP(CDecompyleDiag, CDialog)
	//{{AFX_MSG_MAP(CDecompyleDiag)
	ON_BN_CLICKED(IDC_SELSOURCE, OnSource)
	ON_BN_CLICKED(IDC_SELFILE, OnSelfile)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CDecompyleDiag message handlers

void CDecompyleDiag::OnOK() 
{
	CString sourceFolder;
	GetDlgItem(IDC_SOURCE)->GetWindowText(sourceFolder);

	if (m_listFilename.IsEmpty()) {
		MessageBox(_T("You must select at least one PYC file"), _T("Error"), MB_ICONERROR);
		return;
	}

	// Fetch other data on page
	if (!UpdateData())
		return;


	// Try to start decompylation
	CTime startTime = CTime::GetCurrentTime(), endTime;
	CString filelist = m_listFilename;
	int nbrErr = 0, nbrGood = 0;
	do {
		if (!g_backPython->StartBatch(sourceFolder, filelist, g_config.compareDate, TRUE)) {
			MessageBox(_T("Unable to start decompylation."), _T("Error"), MB_ICONERROR);
			return;
		}

		// Display progress dialog
		CProgress myDiag;
		if (myDiag.DoModal() != IDOK)
			return;		// User cancel operation
		
		// Fetch the result of decompylation
		CStatBatch data;
		g_backPython->GetBatchStatus(data);
		nbrGood += data.curFileNbr;

		if (data.isErr) {
			// Display error message and wait for user answer
			nbrGood--;
			nbrErr++;
			CErrDecomp myDiagErr;
			myDiagErr.m_decompyle = TRUE;
			myDiagErr.m_filename = data.curFilename;
			myDiagErr.m_errMsg = data.errMsg;
			if (myDiagErr.DoModal() != IDOK)
				return;

			// Get rid of last filename and restart decompylation
			filelist = data.filenameToDo;
			int posComma = filelist.Find(_T(','));
			if (posComma >= 0)
				filelist = filelist.Mid(posComma + 1);
			else
				filelist.Empty();

			if (!filelist.IsEmpty())
				data.done = FALSE;			// Continue the while for the next file
		}
		
		if (data.done) {
			endTime = CTime::GetCurrentTime();
			double nbrSec = CTimeSpan(endTime - startTime).GetTotalSeconds();
			CString msg;
			msg.Format(_T("%d files decompile successfully (%d error) in %0.1f minutes."), nbrGood, nbrErr, nbrSec / 60.0);
			MessageBox(msg, _T("Information"), MB_ICONEXCLAMATION);
			break;
		}
	} while (1);

	
	CDialog::OnOK();
}

void CDecompyleDiag::OnSource() 
{
	CString curFolder;
	GetDlgItem(IDC_SOURCE)->GetWindowText(curFolder);

	CFolderSelect mySelect;

	mySelect.m_dirName = curFolder;
	mySelect.m_MsgBrowse = _T("");
	if (mySelect.DoModal(this) == IDOK) {
		g_config.srcFolder = mySelect.m_dirName;
		GetDlgItem(IDC_SOURCE)->SetWindowText(g_config.srcFolder);
	}
}

BOOL CDecompyleDiag::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	GetDlgItem(IDC_SOURCE)->SetWindowText(g_config.srcFolder);

	m_listFilename.Empty();
	DisplayFileList();

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void CDecompyleDiag::DisplayFileList()
{
	CString msg;
	if (m_listFilename.IsEmpty())
		msg = _T("(None)");
	else {
		int nbrFile = CountComma(m_listFilename) + 1;
		if (nbrFile == 1)
			msg = m_listFilename;
		else
			msg.Format(_T("(%d files selected)"), nbrFile);
	}

	GetDlgItem(IDC_FILES)->SetWindowText(msg);
}

void CDecompyleDiag::OnSelfile() 
{
	CSelectFile mySelect;
	GetDlgItem(IDC_SOURCE)->GetWindowText(mySelect.m_sourceDir);
	mySelect.m_fileList = m_listFilename;

	// Make sure there is at least one PYC file
	CFileFind finder;
	CString pattern = mySelect.m_sourceDir + _T("\\*.PYC"); 
	BOOL isFile = finder.FindFile(pattern);
	finder.Close();
	if (!isFile) {
		MessageBox(_T("There is no PYC file in the selected folder."), _T("Error"), MB_ICONERROR);
		return;
	}

	if (mySelect.DoModal() != IDOK)
		return;

	m_listFilename = mySelect.m_fileList;
	DisplayFileList();
}
