// EditFile.cpp : implementation file
//

#include "stdafx.h"
#include "urupython3.h"
#include "EditFile.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CEditFile dialog


CEditFile::CEditFile(CWnd* pParent /*=NULL*/)
	: CDialog(CEditFile::IDD, pParent)
{
	//{{AFX_DATA_INIT(CEditFile)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
}


void CEditFile::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CEditFile)
		// NOTE: the ClassWizard will add DDX and DDV calls here
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CEditFile, CDialog)
	//{{AFX_MSG_MAP(CEditFile)
	ON_EN_CHANGE(IDC_DATA, OnChangeData)
	ON_BN_CLICKED(IDC_GOLINE, OnGoline)
	ON_WM_TIMER()
	ON_BN_CLICKED(IDC_SAVE, OnSave)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CEditFile message handlers

void CEditFile::OnCancel() 
{
	if (m_textChanged) {
		if (MessageBox(_T("Do you want to loose changes?"), _T("Confirmation"), MB_YESNO | MB_DEFBUTTON2 | MB_ICONQUESTION) != IDYES)
			return;
	}
	
	CDialog::OnCancel();
}

void CEditFile::OnOK() 
{
	CWnd *curWin = (CEdit *)GetFocus();
	int idFocus = -1;
	if (curWin)
		idFocus = curWin->GetDlgCtrlID();

	if (idFocus == IDC_LINE) {
		OnGoline();
		return;
	}
}

BOOL CEditFile::OnInitDialog() 
{
	CDialog::OnInitDialog();

	// Read file and put text in edit box
	CFile myFile;
	CFileException errOpen;
	if (!myFile.Open(m_filename, CFile::modeRead | CFile::shareDenyWrite, &errOpen)) {
		CString msg;
		msg.Format(_T("Unable to open file [%s]"), m_filename);
		MessageBox(msg, _T("Error"), MB_ICONERROR);
		EndDialog(IDCANCEL);
		return TRUE;
	}
	
	// Read whole content
	char *buffer = (char *)malloc(myFile.GetLength() + 1);
	if (!buffer) {
		CString msg;
		msg.Format(_T("Out of memory reading file [%s]"), m_filename);
		MessageBox(msg, _T("Error"), MB_ICONERROR);
		EndDialog(IDCANCEL);
		return TRUE;
	}

	int nbrRead = 0;
	try {
		nbrRead = myFile.Read(buffer, myFile.GetLength());

		myFile.Close();
	}
	catch (CFileException *errFile) {
		errFile->Delete();
		try {
			myFile.Close();
		}
		catch (CFileException *errClose) {
			errClose->Delete();
		}

		free(buffer);
		buffer = NULL;

		CString msg;
		msg.Format(_T("Error reading file [%s]"), m_filename);
		MessageBox(msg, _T("Error"), MB_ICONERROR);
		EndDialog(IDCANCEL);
		return TRUE;
	}

	buffer[nbrRead] = 0;		// Make sure at least one null in buffer

	// Make sure that all \n are \n\r
	m_dataOrg = buffer;
	m_dataOrg.Replace(_T("\r\n"), _T("\n"));
	m_dataOrg.Replace(_T("\n"), _T("\r\n"));

	free(buffer);
	buffer = NULL;

	CEdit *editWin = (CEdit *)GetDlgItem(IDC_DATA);
	
	// Set text limits
	UINT limText = UINT(m_dataOrg.GetLength() * 1.10);
	if (limText < 200000)
		limText = 200000;
	editWin->SetLimitText(200000);
	UINT newLim = editWin->GetLimitText();
	if (newLim != limText)
		editWin->SetLimitText(0xffff);		// On Windows 98, limits is 0xffff

	
	// Set font
	m_fixed.CreatePointFont(90, _T("Fixedsys"));
	editWin->SetFont(&m_fixed);

	editWin->SetWindowText(m_dataOrg);
	m_textChanged = FALSE;

	// Read back data from edit box to make sure there is no 'out of memory' problem
	CString dummy;
	editWin->GetWindowText(dummy);
	if (dummy.IsEmpty()) {
		MessageBox(_T("Sorry but the file is too large for this tiny editor.  Use WordPad or another editor."), 
							 _T("Information"), MB_ICONEXCLAMATION);
		EndDialog(IDCANCEL);
		return TRUE;
	}


	// Extract file title
	CString filetitle = m_filename;
	int posLastSlash = filetitle.ReverseFind(_T('\\'));
	if (posLastSlash >= 0)
		filetitle = filetitle.Mid(posLastSlash + 1);

	CString winTitle;
	GetWindowText(winTitle);
	winTitle += _T(" [") + filetitle + _T("]");
	SetWindowText(winTitle);

	GetDlgItem(IDC_LINE)->SetWindowText(_T("1"));

	SetTimer(IDTIMER_EDIT, 250, NULL);

	GetDlgItem(IDC_DATA)->SetFocus();

	return FALSE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}



void CEditFile::OnChangeData() 
{
	m_textChanged = TRUE;	
}

void CEditFile::OnGoline() 
{
	// Get line number
	CString lineStr;
	GetDlgItem(IDC_LINE)->GetWindowText(lineStr);
	lineStr.TrimLeft();
	lineStr.TrimRight();
	if (lineStr.IsEmpty() || (lineStr.SpanIncluding(_T("0123456789")).GetLength() != lineStr.GetLength())) {
		MessageBox(_T("Invalid line number"), _T("Error"), MB_ICONERROR);
		return;
	}
	int lineNumber = _ttoi(lineStr);
	if (lineNumber < 1)
		lineNumber = 1;

	CEdit *editWin = (CEdit *)GetDlgItem(IDC_DATA);
	int maxLine = editWin->GetLineCount();
	if (lineNumber > maxLine)
		lineNumber = maxLine;

	int indexChar = editWin->LineIndex(lineNumber - 1);
	editWin->SetSel(indexChar, indexChar);
	editWin->SetFocus();
}

void CEditFile::OnTimer(UINT nIDEvent) 
{
	if (nIDEvent == IDTIMER_EDIT) {
		CEdit *editWin = (CEdit *)GetFocus();

		if (editWin && (editWin->GetDlgCtrlID() == IDC_DATA)) {
			int curLine = editWin->LineFromChar() + 1;
			CString lineStr;
			lineStr.Format(_T("%d"), curLine);
			GetDlgItem(IDC_LINE)->SetWindowText(lineStr);
		}
	} else
		CDialog::OnTimer(nIDEvent);
}

void CEditFile::OnSave() 
{
	// Save file
	CString newData;
	GetDlgItem(IDC_DATA)->GetWindowText(newData);

	CFile myFile;
	CFileException errOpen;
	if (!myFile.Open(m_filename, CFile::modeCreate | CFile::modeWrite | CFile::shareExclusive, &errOpen)) {
		CString msg;
		msg.Format(_T("Unable to write to file [%s]"), m_filename);
		MessageBox(msg, _T("Error"), MB_ICONERROR);
		return;
	}

	int lenWrite = newData.GetLength();
	char *buffer = newData.GetBuffer(1);
	try {
		myFile.Write(buffer, lenWrite);

		myFile.Close();
	}
	catch (CFileException *errFile) {
		newData.ReleaseBuffer();

		errFile->Delete();
		try {
			myFile.Close();
		}
		catch (CFileException *errClose) {
			errClose->Delete();
		}

		CString msg;
		msg.Format(_T("Error writing to file [%s]"), m_filename);
		MessageBox(msg, _T("Error"), MB_ICONERROR);
		return;
	}

	newData.ReleaseBuffer();

	CDialog::OnOK();
}
