// FilePak.cpp: implementation of the CFilePak class.
//
//////////////////////////////////////////////////////////////////////

// This module handle the PAK file format.  It can pak or unpack it.

#include "stdafx.h"
#include "UruPython3.h"
#include "FilePak.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

CFilePak g_filePak;


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CFilePak::CFilePak()
{

}

CFilePak::~CFilePak()
{

}


// This function unpack a PAK file into the 'destFolder'.  It returns FALSE in case of error and it set the
// 'errMsg' parameter.  Otherwise, it return TRUE and set the 'nbrFile' to the number of file unpacked.  This
// function is based on the first version of UruPython which was a command line application.
BOOL CFilePak::UnpackFile(CString sourceFile, CString destFolder, int &nbrFile, CString &errMsg)
{
	nbrFile = 0;
	errMsg.Empty();

	// Try to read the PAK file in an internal buffer
	FILE *srcHandle = fopen(sourceFile, "rb");
	if (!srcHandle) {
		errMsg = _T("Unable to read source file");
		return FALSE;
	}

	// Get file size
	if (fseek(srcHandle, 0, SEEK_END)) {
		errMsg = _T("Error with fseek (1)");
		fclose(srcHandle);
		return FALSE;
	}
	size_t sourceSize = ftell(srcHandle);
	if (fseek(srcHandle, 0, SEEK_SET)) {
		errMsg = _T("Error with fseek (2)");
		fclose(srcHandle);
		return FALSE;
	}
	if (sourceSize < 16) {
		errMsg = _T("Source file too small");
		fclose(srcHandle);
		return FALSE;
	}

	// Read the whole file in a buffer
	char *buffer = (char *)malloc(sourceSize);
	if (!buffer) {
		errMsg = _T("Out of memory (buffer)");
		fclose(srcHandle);
		return FALSE;
	}
	if (fread(buffer, 1, sourceSize, srcHandle) != sourceSize) {
		errMsg = _T("Read short on source file (buffer)");
		free(buffer);
		fclose(srcHandle);
		return FALSE;
	}

	fclose(srcHandle);

	// Keep the size of the file once decrypted
	long sizeDecrypted = *((long *)(buffer + 12));

	// Check for the 'whatdoyousee' header
	BOOL isOldStyle = FALSE;
	buffer[12] = 0;
	if (strcmp(buffer, "whatdoyousee") != 0) {
		// Check for the 'oldStyle' header

		if (strcmp(buffer, "BriceIsSmart") != 0) {
			errMsg = _T("Not a 'whatdoyousee' or 'BriceIsSmart' file");
			free(buffer);
			return FALSE;
		}

		isOldStyle = TRUE;
	}

	// Decode the file in memory
	size_t offsetData = 0x10;
	while (offsetData < sourceSize) {
		decodeQuad((UINT *)(buffer + offsetData), (UINT *)(buffer + offsetData + 4));
		offsetData += 8;
	}

	// Creates the default python 2.2 header
	char pythonHeader[8];
	pythonHeader[0] = 0x2d;
	pythonHeader[1] = char(0xed);
	pythonHeader[2] = 0x0d;
	pythonHeader[3] = 0x0a;
	*((long *)(pythonHeader + 4)) = time(NULL);

	// Extract file one after the other
	FILE *outHandle;
	char baseName[256], fullName[1024];
	nbrFile = *((int *)(buffer + 16));
	offsetData = 20;
	size_t offsetPYC, sizePYC;
	for (int indexFile = 0; indexFile < nbrFile; indexFile++) {
		// Extract name of PY file
		if (isOldStyle) {
			if (!decodeUruStrOld(buffer + offsetData, baseName)) {
				errMsg = _T("Invalid UruString (old)");
				free(buffer);
				return FALSE;
			}
			offsetData += strlen(baseName) + 4;
		} else {
			if (!decodeUruStr(buffer + offsetData, baseName)) {
				errMsg = _T("Invalid UruString");
				free(buffer);
				return FALSE;
			}
			offsetData += strlen(baseName) + 2;
		}

		// Extract offset of PYC data in buffer
		offsetPYC = *((size_t *)(buffer + offsetData)) + 16;
		offsetData += 4;
		if (offsetPYC > sourceSize) {
			errMsg.Format(_T("Offset too large for [%s]"), baseName);
			free(buffer);
			return FALSE;
		}

		// Extract PYC size located at 'offsetPYC'
		sizePYC = *((size_t *)(buffer + offsetPYC));
		if (offsetPYC + sizePYC > sourceSize) {
			errMsg.Format(_T("Offset+size too large for [%s]"), baseName);
			free(buffer);
			return FALSE;
		}

		// Create output file
		sprintf(fullName, "%s%sc", destFolder, baseName);
		outHandle = fopen(fullName, "wb");
		if (!outHandle) {
			errMsg.Format(_T("unable to create output file [%s]"), fullName);
			free(buffer);
			return FALSE;
		}

		// Write Python header
		if (fwrite(pythonHeader, 1, 8, outHandle) != 8) {
			errMsg.Format(_T("Error writing Python header for [%s]"), fullName);
			fclose(outHandle);
			free(buffer);
			return FALSE;
		}

		// Write PYC data
		if (fwrite(buffer + offsetPYC + 4, 1, sizePYC, outHandle) != sizePYC) {
			errMsg.Format(_T("Error writing Python data for [%s]"), fullName);
			fclose(outHandle);
			free(buffer);
			return FALSE;
		}

		if (fclose(outHandle) != 0) {
			errMsg.Format(_T("Error closing file [%s]"), fullName);
			free(buffer);
			return FALSE;
		}
	}

	free(buffer);

	return TRUE;
}


// Decode two UINT
void CFilePak::decodeQuad(UINT *first, UINT *second)
{
	UINT out1 = *first, out2 = *second, base = 0xc6ef3720;
	UINT temp1, temp2, crossRef[4];

	crossRef[0] = 0x6c0a5452;
	crossRef[1] = 0x03827d0f;
	crossRef[2] = 0x3a170b92;
	crossRef[3] = 0x16db7fc2;

	for (int index = 0x20; index > 0; index--) {
		// First word
		temp1 = temp2 = out1;
		temp1 >>= 5;
		temp2 <<= 4;
		temp1 ^= temp2;
		temp1 += out1;

		temp2 = base;
		temp2 >>= 0x0b;
		temp2 &= 0x03;

		temp2 = crossRef[temp2];
		temp2 += base;
		base += 0x61c88647;

		temp1 ^= temp2;
		out2 -= temp1;

		// Second word
		temp1 = temp2 = out2;
		temp1 >>= 5;
		temp2 <<= 4;
		temp1 ^= temp2;
		temp1 += out2;

		temp2 = base;
		temp2 &= 0x03;

		temp2 = crossRef[temp2];
		temp2 += base;

		temp1 ^= temp2;
		out1 -= temp1;
	}

	*first = out1;
	*second = out2;

}

// Encode two UINT
void CFilePak::encodeQuad(UINT *first, UINT *second)
{
	UINT out1 = *first, out2 = *second, base = 0;
	UINT temp1, temp2, crossRef[4];

	crossRef[0] = 0x6c0a5452;
	crossRef[1] = 0x03827d0f;
	crossRef[2] = 0x3a170b92;
	crossRef[3] = 0x16db7fc2;

	for (int index = 0x20; index > 0; index--) {
		// Second word
		temp1 = temp2 = out2;
		temp1 >>= 5;
		temp2 <<= 4;
		temp1 ^= temp2;
		temp1 += out2;

		temp2 = base;
		temp2 &= 0x03;

		temp2 = crossRef[temp2];
		temp2 += base;
		base -= 0x61c88647;

		temp1 ^= temp2;
		out1 += temp1;

		
		// First word
		temp1 = temp2 = out1;
		temp1 >>= 5;
		temp2 <<= 4;
		temp1 ^= temp2;
		temp1 += out1;

		temp2 = base;
		temp2 >>= 0x0b;
		temp2 &= 0x03;

		temp2 = crossRef[temp2];
		temp2 += base;

		temp1 ^= temp2;
		out2 += temp1;

	}

	*first = out1;
	*second = out2;
}

// This function decode a Uru string (new type which is inverted).  'strEnc' pointed to the encrypted string while
// 'realStr' should point to a buffer of at least 256 bytes (max len for uru string).  Returns FALSE in case of problem.
BOOL CFilePak::decodeUruStr(char *strEnc, char *realStr)
{
	if (!strEnc || !realStr)
		return FALSE;
	if (strEnc[1] != char(0xf0))
		return FALSE;

	int length = strEnc[0];
	memcpy(realStr, strEnc + 2, length);
	realStr[length] = 0;
	for (int index = 0; index < length; index++)
		realStr[index] ^= 0xff;

	return TRUE;
}

// This function create a Uru string (new type which is inverted) based on 'realStr'.  The result is in 'strEnc'.
// Returns FALSE in case of problem.
BOOL CFilePak::encodeUruStr(char *realStr, char *strEnc)
{
	if (!strEnc || !realStr)
		return FALSE;

	int length = strlen(realStr);
	if (length > 255)
		return FALSE;

	strEnc[0] = length & 0xff;
	strEnc[1] = char(0xf0);

	memcpy(strEnc + 2, realStr, length);
	for (int index = 0; index < length; index++)
		strEnc[index + 2] ^= 0xff;

	return TRUE;
}


// This function decode a Uru string (old style which is not inverted and string length on 4 bytes).  'strEnc' pointed
// to the encrypted string while 'realStr' should point to a buffer of at least 256 bytes (normally, it should be up to
// 65536 but I limit it to 256 in thid function). Returns FALSE in case of problem.
BOOL CFilePak::decodeUruStrOld(char *strEnc, char *realStr)
{
	if (!strEnc || !realStr)
		return FALSE;

	int length = *((int *)strEnc);
	if (length > 255)
		return FALSE;

	memcpy(realStr, strEnc + 4, length);
	realStr[length] = 0;

	return TRUE;
}


// Create a PAK from a list of PYC files.  'destFile' is the complete name of the PAK file that will be created.
// 'sourceFolder' is the forlder in which the PYC are stored.  'listFile' is a comma separated list of file
// (no extension, assumed PYC).  'errMsg' will be set to the cause of the problem if function returns FALSE.
// Otherwise, it return the number of file stored in the PAK file in 'nbrFile' (it's only an output variable)
// This function is derived from a console program and I've changed the minimum to adapt it to an MFC program.
BOOL CFilePak::pakPython(CString destFile, CString sourceFolder, CString listFile, int &nbrFile, CString &errMsg)
{
	errMsg.Empty();
	nbrFile = 0;

	if (destFile.IsEmpty() || listFile.IsEmpty()) {
		errMsg = _T("Invalid dirName or destName");
		return FALSE;
	}

	// First, count the number of PYC file and the size of each of them
	CString tempList = listFile;
	CString curFilename, fullName;
	CFileFind finder;
	DWORD totalSize = 0;		// Total size not counting the whatdouyousee header (16 bytes)
	DWORD sizeHeader = 4;		// Will be the size of the list of filename with offset (include at least the number of file)

	do {
		// Extract next filename
		int posComma = tempList.Find(_T(','));
		if (posComma >= 0)
			curFilename = tempList.Left(posComma);
		else
			curFilename = tempList;

		curFilename.TrimLeft();
		curFilename.TrimRight();
		fullName = sourceFolder + curFilename + _T(".pyc");

		if (finder.FindFile(fullName)) {
			// Check that it is a normal file
			finder.FindNextFile();
			if (!(finder.IsNormal() || finder.IsArchived())) {
				errMsg = _T("Not a 'normal' file (system or hidden or directory) [") + curFilename + _T("]");
				return FALSE;
			}
			nbrFile++;

			// Will need space to store the name of the file plus ".PY".  The Uru string then add 2 bytes.  I then
			// need 4 bytes for the offset
			sizeHeader += strlen(curFilename) + 3 + 2 + 4;

			// The file will be stored completely except for the 4 magic number (the timestamp will be replaced by file size)
			totalSize += finder.GetLength() - 4;
		} else {
			errMsg = _T("File not found [") + fullName + _T("]");
			return FALSE;
		}
		finder.Close();

		// Extract from the list of file the last filename processed 
		if (posComma >= 0)
			tempList = tempList.Mid(posComma + 1);
		else
			tempList.Empty();

		tempList.TrimLeft();
		tempList.TrimRight();
	} while (!tempList.IsEmpty());

	if (nbrFile < 1) {
		errMsg = _T("No PYC file to store");
		return FALSE;
	}
	totalSize += sizeHeader;

	// Adjust output size to be on an 8 bytes boundary (for the encryption) and add the size of the whatdoyousee header
	size_t totalBuffer = totalSize + 16;
	int diffBound = totalBuffer % 8;
	if (diffBound)
		totalBuffer += (8 - diffBound);

	// Creates a buffer to store all data in memory
	char *buffer = (char *)malloc(totalBuffer);
	if (!buffer) {
		errMsg = _T("Out of memory");
		return FALSE;
	}

	// Store header information
	strcpy(buffer, "whatdoyousee");
	*((long *)(buffer + 12)) = totalSize;
	*((long *)(buffer + 16)) = nbrFile;

	// Now scan the file again and stored them into buffer
	int newNbrFile = 0;		// Just to double check the number of files
	size_t offsetHeader = 16 + 4;
	size_t offsetData = offsetHeader + sizeHeader - 4;		// need to get rid of the +4 in offsetHeader which is also in sizeHeader

	FILE *srcHandle;
	char shortName[256];

	tempList = listFile;
	do {
		// Extract next filename
		int posComma = tempList.Find(_T(','));
		if (posComma >= 0)
			curFilename = tempList.Left(posComma);
		else
			curFilename = tempList;

		curFilename.TrimLeft();
		curFilename.TrimRight();
		fullName = sourceFolder + curFilename + _T(".pyc");

		if (!finder.FindFile(fullName)) {
			errMsg = _T("File not found [") + fullName + _T("]");
			free(buffer);
			return FALSE;
		}

		// Check that it is a normal file
		finder.FindNextFile();
		if (!(finder.IsNormal() || finder.IsArchived())) {
			errMsg = _T("Not a 'normal' file (system or hidden or directory) [") + curFilename + _T("]");
			return FALSE;
		}
		newNbrFile++;

		// Stored filename in header
		strcpy(shortName, finder.GetFileName());
		shortName[strlen(shortName) - 1] = 0;			// Get rid of the C in PYC
		if (!encodeUruStr(shortName, buffer + offsetHeader)) {
			errMsg = _T("Error with encodeUruStr [") + curFilename + _T("]");
			free(buffer);
			return FALSE;
		}
		offsetHeader += strlen(shortName) + 2;
		
		// Store file offset in the header
		*((long *)(buffer + offsetHeader)) = offsetData - 16;			// The -16 is for the size of the 'whatdoyousee' header
		offsetHeader += 4;

		// Check for header overflow in the data section (the +16 if for the whatdoyousee header)
		if (offsetHeader > (sizeHeader + 16)) {
			errMsg.Format(_T("Header buffer overflow in data buffer at file [%s]"), curFilename);
			free(buffer);
			return FALSE;
		}

		// Open source file
		srcHandle = fopen(fullName, "rb");
		if (!srcHandle) {
			errMsg.Format(_T("Error opening file [%s] %d"), curFilename, errno);
			free(buffer);
			return FALSE;
		}

		// Jump over the magic number (keep timestamp but will be changed to file size)
		if (fseek(srcHandle, 4, SEEK_SET) != 0) {
			errMsg.Format(_T("Error with seek for file [%s]"), curFilename);
			fclose(srcHandle);
			free(buffer);
			return FALSE;
		}
		
		// Check for buffer overflow (-4 because I don't read magic number and -16 because of the whatdoyousee header)
		if ((offsetData + finder.GetLength() - 4 - 16) > totalSize) {
			errMsg.Format(_T("Data buffer overflow at file [%s]"), curFilename);
			fclose(srcHandle);
			free(buffer);
			return FALSE;
		}

		if (fread(buffer + offsetData, 1, finder.GetLength() - 4, srcHandle) != (finder.GetLength() - 4)) {
			errMsg.Format(_T("Short read for file [%s]"), curFilename);
			fclose(srcHandle);
			free(buffer);
			return FALSE;
		}
		fclose(srcHandle);

		// Change timestamp to file size
		*((long *)(buffer + offsetData)) = finder.GetLength() - 4 - 4;		// Don`t count the magic number and the timestamp
		offsetData += finder.GetLength() - 4;
		
		finder.Close();

		// Extract from the list of file the last filename processed 
		if (posComma >= 0)
			tempList = tempList.Mid(posComma + 1);
		else
			tempList.Empty();

		tempList.TrimLeft();
		tempList.TrimRight();
	} while (!tempList.IsEmpty());

	// Make sure we have stored the correct number of file
	if (newNbrFile != nbrFile) {
		errMsg = _T("Number of PYC file mismatch");
		free(buffer);
		return FALSE;
	}
	if (offsetData != (totalSize + 16)) {
		errMsg = _T("PYC file size mismatch");
		free(buffer);
		return FALSE;
	}

	// Encrypt the buffer
	offsetData = 0x10;
	while (offsetData < totalBuffer) {
		encodeQuad((UINT *)(buffer + offsetData), (UINT *)(buffer + offsetData + 4));
		offsetData += 8;
	}

	// Save buffer to disk
	FILE *destHandle = fopen(destFile, "wb");
	if (!destHandle) {
		errMsg = _T("Error creating output file");
		free(buffer);
		return FALSE;
	}

	if (fwrite(buffer, 1, totalBuffer, destHandle) != totalBuffer) {
		errMsg = _T("Error writing output file");
		fclose(destHandle);
		free(buffer);
		return FALSE;
	}

	if (fclose(destHandle) != 0) {
		errMsg = _T("Error closing output file");
		free(buffer);
		return FALSE;
	}

	free(buffer);

	return TRUE;
}

