// Python.cpp: implementation of the CPython class.
//
//////////////////////////////////////////////////////////////////////

// This module encapsulate the CYPYTHON22.DLL functionnality

#include "stdafx.h"
#include "UruPython3.h"
#include "Python.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif



//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPython::CPython()
{
	m_hDLL = NULL;

	Py_Initialize = NULL;
	Py_IsInitialized = NULL;
	Py_Finalize = NULL;
	PyRun_SimpleString = NULL;
	Py_SetProgramName = NULL;
	PyImport_ImportModule = NULL;
	PyModule_GetDict = NULL;
	PyDict_SetItemString = NULL;
	PyObject_GetAttrString = NULL;
	PyCObject_AsVoidPtr = NULL;
	PyObject_AsCharBuffer = NULL;
	PyObject_CallMethod = NULL;

	m_modcStringIO = NULL;
	m_attrStr = NULL;
	m_objCAPI = NULL;
	m_newBufObj = NULL;

}

CPython::~CPython()
{
	CloseLink();
}

// Close link with the python object.  Freeing (almost...) all memory used up.
// Just to fact of doing a Initilized and a Finalize will leak around 24 K of memory!
// Not so bad for a hack!
void CPython::CloseLink()
{
	// Free memory of every object created by the python interpreter
	if (m_newBufObj) {
		Py_DECREF(m_newBufObj);
		m_newBufObj = NULL;
	}
	if (m_attrStr) {
		Py_DECREF(m_attrStr);
		m_attrStr = NULL;
	}
	if (m_modcStringIO) {
		Py_DECREF(m_modcStringIO);
		m_modcStringIO = NULL;
	}

	// Try to terminate python interpreter
	if (Py_IsInitialized && Py_Finalize) {
		if (Py_IsInitialized())
			Py_Finalize();
	}

	// Close link with DLL
	if (m_hDLL) {
		FreeLibrary(m_hDLL);
		m_hDLL = NULL;
	}

	Py_Initialize = NULL;
	Py_IsInitialized = NULL;
	Py_Finalize = NULL;
	PyRun_SimpleString = NULL;
	Py_SetProgramName = NULL;
	PyImport_ImportModule = NULL;
	PyModule_GetDict = NULL;
	PyDict_SetItemString = NULL;
	PyObject_GetAttrString = NULL;
	PyCObject_AsVoidPtr = NULL;
	PyObject_AsCharBuffer = NULL;
	PyObject_CallMethod = NULL;
}


// Create the link with the python DLL.  Returns TRUE upon success.  Otherwise, the 'errMsg'
// is set to the cause of the error.
BOOL CPython::OpenLink(CString &errMsg)
{
	CloseLink();

	errMsg.Empty();

	// Python interpreter always open its PAK file from the Python directory in the current directory so
	// set it properly
	CString pathPAK = g_config.appPath;
	while ((pathPAK.GetLength() > 1) && (pathPAK[pathPAK.GetLength() - 1] == _T('\\')))
		pathPAK = pathPAK.Left(pathPAK.GetLength() - 1);
	SetCurrentDirectory(pathPAK);

	// First link with the DLL
	CString filename  = g_config.appPath + g_config.pythonDLL;
	m_hDLL = LoadLibrary(filename);
	if (!m_hDLL) {
		errMsg.Format(_T("Error with LoadLibrary: %ul"), GetLastError());
		return FALSE;
	}

	// Creates a link with every function needed in the DLL
	Py_Initialize = (Py_InitializeType)GetProcAddress(m_hDLL, "Py_Initialize");
	if (!Py_Initialize) {
		errMsg.Format(_T("Error with GetProcAddress(Py_Initialize): %ul"), GetLastError());
		CloseLink();
		return FALSE;
	}

	Py_IsInitialized = (Py_IsInitializedType)GetProcAddress(m_hDLL, "Py_IsInitialized");
	if (!Py_IsInitialized) {
		errMsg.Format(_T("Error with GetProcAddress(Py_IsInitialized): %ul"), GetLastError());
		CloseLink();
		return FALSE;
	}

	Py_Finalize = (Py_FinalizeType)GetProcAddress(m_hDLL, "Py_Finalize");
	if (!Py_Finalize) {
		errMsg.Format(_T("Error with GetProcAddress(Py_Finalize): %ul"), GetLastError());
		CloseLink();
		return FALSE;
	}

	PyRun_SimpleString = (PyRun_SimpleStringType)GetProcAddress(m_hDLL, "PyRun_SimpleString");
	if (!PyRun_SimpleString) {
		errMsg.Format(_T("Error with GetProcAddress(PyRun_SimpleString): %ul"), GetLastError());
		CloseLink();
		return FALSE;
	}

	Py_SetProgramName = (Py_SetProgramNameType)GetProcAddress(m_hDLL, "Py_SetProgramName");
	if (!Py_SetProgramName) {
		errMsg.Format(_T("Error with GetProcAddress(Py_SetProgramName): %ul"), GetLastError());
		CloseLink();
		return FALSE;
	}

	PyImport_ImportModule = (PyImport_ImportModuleType)GetProcAddress(m_hDLL, "PyImport_ImportModule");
	if (!PyImport_ImportModule) {
		errMsg.Format(_T("Error with GetProcAddress(PyImport_ImportModule): %ul"), GetLastError());
		CloseLink();
		return FALSE;
	}

	PyModule_GetDict = (PyModule_GetDictType)GetProcAddress(m_hDLL, "PyModule_GetDict");
	if (!PyModule_GetDict) {
		errMsg.Format(_T("Error with GetProcAddress(PyModule_GetDict): %ul"), GetLastError());
		CloseLink();
		return FALSE;
	}

	PyDict_SetItemString = (PyDict_SetItemStringType)GetProcAddress(m_hDLL, "PyDict_SetItemString");
	if (!PyDict_SetItemString) {
		errMsg.Format(_T("Error with GetProcAddress(PyDict_SetItemString): %ul"), GetLastError());
		CloseLink();
		return FALSE;
	}

	PyObject_GetAttrString = (PyObject_GetAttrStringType)GetProcAddress(m_hDLL, "PyObject_GetAttrString");
	if (!PyObject_GetAttrString) {
		errMsg.Format(_T("Error with GetProcAddress(PyObject_GetAttrString): %ul"), GetLastError());
		CloseLink();
		return FALSE;
	}

	PyCObject_AsVoidPtr = (PyCObject_AsVoidPtrType)GetProcAddress(m_hDLL, "PyCObject_AsVoidPtr");
	if (!PyCObject_AsVoidPtr) {
		errMsg.Format(_T("Error with GetProcAddress(PyCObject_AsVoidPtr): %ul"), GetLastError());
		CloseLink();
		return FALSE;
	}

	PyObject_AsCharBuffer = (PyObject_AsCharBufferType)GetProcAddress(m_hDLL, "PyObject_AsCharBuffer");
	if (!PyObject_AsCharBuffer) {
		errMsg.Format(_T("Error with GetProcAddress(PyObject_AsCharBuffer): %ul"), GetLastError());
		CloseLink();
		return FALSE;
	}

	PyObject_CallMethod = (PyObject_CallMethodType)GetProcAddress(m_hDLL, "PyObject_CallMethod");
	if (!PyObject_CallMethod) {
		errMsg.Format(_T("Error with GetProcAddress(PyObject_CallMethod): %ul"), GetLastError());
		CloseLink();
		return FALSE;
	}

	// Now do the initialization of the python interpreter the same way as Uru normally do (as usual...almost...)
	Py_SetProgramName(_T("plasma"));

	Py_Initialize();
	if (!Py_IsInitialized()) {
		errMsg.Format(_T("Unable to initialize python object"));
		CloseLink();
		return FALSE;
	}

	// Initialized the object use to fetch the stderr and stdout from the python interpreter
	m_modcStringIO = PyImport_ImportModule(_T("cStringIO"));
	if (!m_modcStringIO) {
		errMsg.Format(_T("Error with PyImport_ImportModule(cStringIO)"));
		CloseLink();
		return FALSE;
	}
	m_attrStr = PyObject_GetAttrString(m_modcStringIO, _T("cStringIO_CAPI"));
	if (!m_attrStr) {
		errMsg.Format(_T("Error with PyObject_GetAttrString(cStringIO_CAPI)"));
		CloseLink();
		return FALSE;
	}

	m_objCAPI = (PycStringIO_CAPI *)PyCObject_AsVoidPtr(m_attrStr);
	if (!m_objCAPI) {
		errMsg.Format(_T("Error with PyCObject_AsVoidPtr"));
		CloseLink();
		return FALSE;
	}
	
	m_newBufObj = m_objCAPI->NewOutput(60000);
	if (!m_newBufObj) {
		errMsg.Format(_T("Error with m_objCAPI->NewOutput(60000)"));
		CloseLink();
		return FALSE;
	}

	PyObject *modSys = PyImport_ImportModule(_T("sys"));
	if (!modSys) {
		errMsg.Format(_T("Error with PyImport_ImportModule(sys)"));
		CloseLink();
		return FALSE;
	}
	
	PyObject *dictSys = PyModule_GetDict(modSys);
	if (!dictSys) {
		errMsg.Format(_T("Error with PyModule_GetDict(sys)"));
		CloseLink();
		return FALSE;
	}

	if (PyDict_SetItemString(dictSys, _T("stdout"), m_newBufObj) < 0) {
		Py_DECREF(dictSys);
		Py_DECREF(modSys);

		errMsg.Format(_T("Error with PyDict_SetItemString(stdout)"));
		CloseLink();
		return FALSE;
	}

	if (PyDict_SetItemString(dictSys, _T("stderr"), m_newBufObj) < 0) {
		Py_DECREF(dictSys);
		Py_DECREF(modSys);

		errMsg.Format(_T("Error with PyDict_SetItemString(stderr)"));
		CloseLink();
		return FALSE;
	}

	// Free up all local object
	Py_DECREF(dictSys);
	//Py_DECREF(modSys);		// Strange...cannot finalize python if I get rid of the reference???

	// Python interpreter created succesfully
	return TRUE;
}


// Decrement a reference on a python object (very crude version...)
void CPython::Py_DECREF(PyObject *objPtr)
{
	if (--(objPtr)->ob_refcnt <= 0) {
		typedef void (*destructor)(PyObject *);

		typedef void (__cdecl* DestructorType)(PyObject *);
		DestructorType destrucFunc = (DestructorType)*((int *)(objPtr->object + 0x18));
		destrucFunc(objPtr);
	}
}

// Returns TRUE if python interpreter is active
BOOL CPython::IsInit()
{
	if (m_hDLL && Py_IsInitialized)
		return(Py_IsInitialized());
	else
		return FALSE;
}

// Returns the content of the stdout and stderr of the python interpreter
CString CPython::GetOutput()
{
	CString data;
	if (IsInit() && m_objCAPI && m_newBufObj) {
		PyObject *bufferOut = m_objCAPI->cgetvalue(m_newBufObj);
		if (bufferOut) {
			char *dataBuf;
			int lenDataBuf;

			if (PyObject_AsCharBuffer(bufferOut, &dataBuf, &lenDataBuf) == 0)
				data = dataBuf;

			Py_DECREF(bufferOut);

			PyObject *refReset = PyObject_CallMethod(m_newBufObj, _T("reset"), NULL);
			if (refReset)
				Py_DECREF(refReset);
		}
	}

	return data;
}


// Try to decompile the 'filenameIn' into the 'filenameOut'.  Both filename must contain the
// whole path.  Return FALSE if it cannot execute the command.  'msg' is set to either the cause
// of the error or the ouput buffer of the python interpreter.
BOOL CPython::Decompyle(CString filenameIn, CString filenameOut, CString &msg)
{
	if (!IsInit() || !PyRun_SimpleString) {
		msg = _T("Python interpreter not initialized");
		return FALSE;
	}

	// Format filename properly for Python (get rid of backslash and add backslash before quote)
	CString pythonIn = filenameIn, pythonOut = filenameOut;
	pythonIn.Replace(_T('\\'), _T('/'));
	pythonIn.Replace(_T("'"), _T("\\'"));

	pythonOut.Replace(_T('\\'), _T('/'));
	pythonOut.Replace(_T("'"), _T("\\'"));


	// Delete filenameOut to be sure that python will recreate it
	DeleteFile(filenameOut);

	CString cmd;
	cmd.Format(_T("import decompyle,sys\n")
						 _T("outHandle = open('%s', 'w')\n")
						 _T("decompyle.decompyle_file('%s', outHandle)\n")
						 _T("outHandle.close()\n"),
						 pythonOut, pythonIn);

	
	if (PyRun_SimpleString(cmd) < 0) {
		msg = _T("Error with PyRun_SimpleString\r\n") + GetOutput();
		return FALSE;
	}

	// Check if the file has been created
	CFileFind finder;
	BOOL isFile = finder.FindFile(filenameOut);
	DWORD fileLen = 0;
	if (isFile) {
		finder.FindNextFile();
		fileLen = finder.GetLength();
	}
	finder.Close();

	if (!isFile || (fileLen < 8)) {
		msg = _T("Out file not created\r\n") + GetOutput();
		return FALSE;
	}

	// Check content of output buffer to see any error
	msg = GetOutput();
	if (msg.Find(_T("Traceback")) >= 0) {
		return FALSE;
	}

	return TRUE;
}


// Try to compile the 'filenameIn' (will create same name with a 'c' appended).  Filename must contain the
// whole path.  Return FALSE if it cannot execute the command.  'msg' is set to either the cause
// of the error or the ouput buffer of the python interpreter.
BOOL CPython::Compile(CString filenameIn, CString &msg)
{
	if (!IsInit() || !PyRun_SimpleString) {
		msg = _T("Python interpreter not initialized");
		return FALSE;
	}

	// Format filename properly for Python (get rid of backslash and add backslash before quote)
	CString pythonIn = filenameIn;
	pythonIn.Replace(_T('\\'), _T('/'));
	pythonIn.Replace(_T("'"), _T("\\'"));

	// Delete filenameOut to be sure that python will recreate it
	CString filenameOut = filenameIn + _T('c');
	DeleteFile(filenameOut);

	CString cmd;
	cmd.Format(_T("import py_compile\n")
						 _T("py_compile.compile('%s')\n"),
						 pythonIn);

	
	if (PyRun_SimpleString(cmd) < 0) {
		msg = _T("Error with PyRun_SimpleString\r\n") + GetOutput();
		return FALSE;
	}

	// Check if the file has been created
	CFileFind finder;
	BOOL isFile = finder.FindFile(filenameOut);
	DWORD fileLen = 0;
	if (isFile) {
		finder.FindNextFile();
		fileLen = finder.GetLength();
	}
	finder.Close();

	if (!isFile || (fileLen < 1)) {
		msg = _T("Out file not created\r\n") + GetOutput();
		return FALSE;
	}

	// Check content of output buffer to see any error
	msg = GetOutput();
	if (msg.Find(_T("Traceback")) >= 0) {
		return FALSE;
	}

	return TRUE;
}

